// Copyright 2016 Canonical Ltd.
// Licensed under the AGPLv3, see LICENCE file for details.

package stateenvirons

import (
	"github.com/juju/errors"
	"gopkg.in/juju/names.v3"

	"github.com/juju/juju/caas"
	"github.com/juju/juju/cloud"
	"github.com/juju/juju/environs"
	"github.com/juju/juju/environs/config"
	"github.com/juju/juju/state"
)

// Model exposes the methods needed for an EnvironConfigGetter.
type Model interface {
	ModelTag() names.ModelTag
	ControllerUUID() string
	Type() state.ModelType
	Cloud() (cloud.Cloud, error)
	CloudRegion() string
	CloudCredential() (state.Credential, bool, error)
	Config() (*config.Config, error)
}

// EnvironConfigGetter implements environs.EnvironConfigGetter
// in terms of a Model.
type EnvironConfigGetter struct {
	Model

	// NewContainerBroker is a func that returns a caas container broker
	// for the relevant model.
	NewContainerBroker caas.NewContainerBrokerFunc
}

// CloudAPIVersion returns the cloud API version for the cloud with the given spec.
func (g EnvironConfigGetter) CloudAPIVersion(spec environs.CloudSpec) (string, error) {
	// Only CAAS models have an API version we care about right now.
	if g.Model.Type() == state.ModelTypeIAAS {
		return "", nil
	}
	cfg, err := g.Config()
	if err != nil {
		return "", errors.Trace(err)
	}
	newBroker := g.NewContainerBroker
	if newBroker == nil {
		newBroker = caas.New
	}
	broker, err := newBroker(environs.OpenParams{
		ControllerUUID: g.Model.ControllerUUID(),
		Cloud:          spec,
		Config:         cfg,
	})
	if err != nil {
		return "", errors.Trace(err)
	}
	return broker.APIVersion()
}

// ModelConfig implements environs.EnvironConfigGetter.
func (g EnvironConfigGetter) ModelConfig() (*config.Config, error) {
	return g.Config()
}

// CloudSpec implements environs.EnvironConfigGetter.
func (g EnvironConfigGetter) CloudSpec() (environs.CloudSpec, error) {
	cloud, err := g.Model.Cloud()
	if err != nil {
		return environs.CloudSpec{}, errors.Trace(err)
	}
	regionName := g.Model.CloudRegion()
	credentialValue, ok, err := g.Model.CloudCredential()
	if err != nil {
		return environs.CloudSpec{}, errors.Trace(err)
	}
	var credential *state.Credential
	if ok {
		credential = &credentialValue
	}
	return CloudSpec(cloud, regionName, credential)
}

// CloudSpec returns an environs.CloudSpec from a *state.State,
// given the cloud, region and credential names.
func CloudSpec(
	modelCloud cloud.Cloud,
	regionName string,
	credential *state.Credential,
) (environs.CloudSpec, error) {
	var cloudCredential *cloud.Credential
	if credential != nil {
		cloudCredentialValue := cloud.NewNamedCredential(credential.Name,
			cloud.AuthType(credential.AuthType),
			credential.Attributes,
			credential.Revoked,
		)
		cloudCredential = &cloudCredentialValue
	}

	return environs.MakeCloudSpec(modelCloud, regionName, cloudCredential)
}

// NewEnvironFunc defines the type of a function that, given a state.State,
// returns a new Environ.
type NewEnvironFunc func(Model) (environs.Environ, error)

// GetNewEnvironFunc returns a NewEnvironFunc, that constructs Environs
// using the given environs.NewEnvironFunc.
func GetNewEnvironFunc(newEnviron environs.NewEnvironFunc) NewEnvironFunc {
	return func(m Model) (environs.Environ, error) {
		g := EnvironConfigGetter{Model: m}
		return environs.GetEnviron(g, newEnviron)
	}
}

// NewCAASBrokerFunc defines the type of a function that, given a state.State,
// returns a new CAAS broker.
type NewCAASBrokerFunc func(Model) (caas.Broker, error)

// GetNewCAASBrokerFunc returns a NewCAASBrokerFunc, that constructs CAAS brokers
// using the given caas.NewContainerBrokerFunc.
func GetNewCAASBrokerFunc(newBroker caas.NewContainerBrokerFunc) NewCAASBrokerFunc {
	return func(m Model) (caas.Broker, error) {
		g := EnvironConfigGetter{Model: m}
		cloudSpec, err := g.CloudSpec()
		if err != nil {
			return nil, errors.Trace(err)
		}
		cfg, err := g.Config()
		if err != nil {
			return nil, errors.Trace(err)
		}
		return newBroker(environs.OpenParams{
			ControllerUUID: m.ControllerUUID(),
			Cloud:          cloudSpec,
			Config:         cfg,
		})
	}
}
